package com.oilabs.map.insuideindoor.ar.ar;

import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;

import com.oilabs.map.insuideindoor.ar.ar.wrapper.Api;
import com.indooratlas.android.sdk.IAARObject;
import com.indooratlas.android.sdk.IAARSession;
import com.indooratlas.android.sdk.IALocation;
import com.indooratlas.android.sdk.IALocationListener;
import com.indooratlas.android.sdk.IALocationManager;
import com.indooratlas.android.sdk.IALocationRequest;
import com.indooratlas.android.sdk.IAPOI;
import com.indooratlas.android.sdk.IARegion;
import com.indooratlas.android.sdk.IAWayfindingRequest;
import com.indooratlas.android.sdk.resources.IALatLng;
import com.indooratlas.android.sdk.resources.IAVenue;

import org.json.JSONException;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.List;

import timber.log.Timber;

public class WayfindingSession implements IALocationListener, IARegion.Listener {
    private final IALocationManager locationManager;
    private final IAARSession arSdk;
    private final Callbacks callbacks;
    private final Handler _handler = new Handler(Looper.getMainLooper());
    private final List<IAPOI> iapois;

    interface Callbacks {
        void onArPoisUpdated(List<ArPOI> arPois);
    }

    public static class ArPOI {
        IAPOI iaPoi;
        double heading;
        double elevation;
        String textureName;
        float scale = 1f;
    }

    WayfindingSession(IALocationManager manager, Callbacks cbs, List<IAPOI> iapois) {
        this.callbacks = cbs;

        locationManager = manager;
        this.iapois = iapois;
        arSdk = locationManager.requestArUpdates();

        locationManager.requestLocationUpdates(IALocationRequest.create(), this);
        locationManager.registerRegionListener(this);
        _handler.postDelayed(() -> {
            locationManager.setLocation(new IALocation.Builder()
                    .withLatitude(lat)
                    .withLongitude(lng)
                    .withAccuracy(10)
                    .withFloorLevel(1)
                    .build()
            );
            //double lt = sLat+(cLat - lat), lg = sLng + (cLng - lng);
            Timber.tag("FAKELOC").d("%s, %s", lat, lng);
        }, 2000);
    }

    void destroy() {
        arSdk.destroy();
        locationManager.removeWayfindingUpdates();
        locationManager.unregisterRegionListener(this);
        locationManager.removeLocationUpdates(this);
    }

    boolean converged() {
        //return true;
        return arSdk.converged();
    }

    void onArFrame(float[] imuToWorld, float[] invViewMat, List<Api.HorizontalPlane> planes) {
        arSdk.setPoseMatrix(imuToWorld);
        arSdk.setCameraToWorldMatrix(invViewMat);
        for (Api.HorizontalPlane plane : planes) {
            arSdk.addArPlane(
                    plane.xyz,
                    plane.extentX,
                    plane.extentZ);
        }
    }

    IAARSession getArSdk() {
        return arSdk;
    }

    IALocationManager getLocationManager() {
        return locationManager;
    }

    List<IAARObject> getWaypoints() {
        return arSdk.getWayfindingTurnArrows();
    }

    public void setWayfindingTarget(IAWayfindingRequest wayfindingRequest) {
        if (wayfindingRequest != null) {
            Timber.i("Setting AR wayfinding target to %f,%f,%d",
                    wayfindingRequest.getLatitude(),
                    wayfindingRequest.getLongitude(),
                    wayfindingRequest.getFloor());
            arSdk.startWayfinding(wayfindingRequest);
        }
    }

    @Override
    public void onEnterRegion(IARegion region) {
        Timber.d("ENTER region %s", region.getName());
        IAVenue venue = region.getVenue();
        if (venue != null) {
            Timber.i("ENTER venue %s", venue.getName());
            // update POIs
            List<ArPOI> newArPois = new ArrayList<>();
            int i = 0;
            List<IAPOI> iapois = this.iapois;
            for (IAPOI poi : iapois) {
                ArPOI arPoi = new ArPOI();
                double elevation = 0.5;
                // use one FP by random to simplify -- usually they are the same for other
                // could perhaps match a bit more accurately with POI.floornumber --> fp.bearing
                double heading = venue.getFloorPlans().get(0).getBearing();
                Timber.d("POI " + poi.getId() + ", heading: " + heading);
                elevation = 1;
                // use a default image for all POIS in Camera view -- useful for quick demos
                arPoi.textureName = "IA_AR_ad_framed.png";
                arPoi.scale = 1;

                if (poi.getPayload() != null) {
                    try {
                        Timber.d(poi.getPayload().toString());
                        JSONObject arPayload = poi.getPayload().getJSONObject("");
                        if (arPayload == null) {
                            Timber.d("POI " + poi.getId() + " does not have AR payload, " +
                                    "using default IA image");
                        } else {
                            elevation = arPayload.getDouble("elevation");
                            heading = arPayload.getDouble("heading");
                            // NOTE: a potential security issue here. Should validate that the asset
                            // exists and is not pointing to an illegal path here
                            arPoi.textureName = arPayload.getString("assetName");
                            arPoi.scale = (float) arPayload.getDouble("scale");
                        }
                    } catch (JSONException ex) {
                        Timber.e("failed to parse AR payload, use default IA image");
                    }
                }

                arPoi.iaPoi = poi;
                arPoi.heading = heading;
                arPoi.elevation = elevation;

                newArPois.add(arPoi);
            }

            Timber.d("%s AR POI(s)", newArPois.size());
            callbacks.onArPoisUpdated(newArPois);
        }
    }

    private List<IAPOI> getPoIs() {
        List<IAPOI> pois = new ArrayList<>();
        pois.add(new IAPOI("1", 1, new IALatLng(19.09614069661849,72.87494458659683), "Burger King", null));
        pois.add(new IAPOI("2", 1, new IALatLng(19.09930713759286,72.87373404109417), "Shiv Sagar", null));
        pois.add(new IAPOI("3", 1, new IALatLng(19.09646722213212,72.8744603204529), "Irani Cafe", null));
        return pois;
    }

    @Override
    public void onExitRegion(IARegion region) {
    }

    // TODO: should not be needed
    @Override
    public void onLocationChanged(IALocation location) {
        cLat = location.getLatitude();
        cLng = location.getLongitude();

        //Timber.tag("FAKELOC").d("%s, %s", lt, lg);
    }

    @Override
    public void onStatusChanged(String provider, int status, Bundle extras) {
    }

    private final double lat = 19.098322248385287;
    private final double lng = 72.8746381639334;
    private double cLat = 0;
    private double cLng = 0;
    private final double oLat = 26.2525967;
    private final double oLng = 72.9984262;
}
