package com.oilabs.map.insuideindoor.wayfinding

import android.content.Context
import com.oilabs.map.insuideindoor.utils.ChannelData
import com.indooratlas.android.sdk.*
import com.indooratlas.android.sdk.resources.IAFloorPlan
import com.indooratlas.android.sdk.resources.IALatLng
import com.indooratlas.android.sdk.resources.IALocationListenerSupport
import com.indooratlas.android.sdk.resources.IAVenue
import io.flutter.Log

object IAWayfindingImpl :
    IAWayfinding {

    private val TAG = "IA_WAYFINDING"

    private var mFloor = 0
    private var cameraPositionNeedsUpdating = true
    private var mCurrentFloorPlan: IARegion? = null
    private lateinit var mVenue: IAVenue
    private var mCurrentRoute: IARoute? = null
    private var mWayfindingDestination: IAWayfindingRequest? = null
    private lateinit var mIALocationManager: IALocationManager
    private lateinit var sendDataToFlutter: (ChannelData) -> Unit


    override fun setUpIAWayfinding(
        context: Context,
        sendDataToFlutter: (ChannelData) -> Unit
    ) {
        this.sendDataToFlutter = sendDataToFlutter
        mIALocationManager = IALocationManager.create(context)
        mIALocationManager.requestLocationUpdates(IALocationRequest.create(), mListener)
        mIALocationManager.registerRegionListener(mRegionListener)
        mIALocationManager.registerOrientationListener(
            IAOrientationRequest(1.0, 0.0),
            mOrientationListener
        )
    }

    override fun onResume() {
        mIALocationManager.requestLocationUpdates(IALocationRequest.create(), mListener)
        mIALocationManager.registerRegionListener(mRegionListener)
        mIALocationManager.registerOrientationListener(
            IAOrientationRequest(1.0, 0.0),
            mOrientationListener
        )

        if (mWayfindingDestination != null) {
            mIALocationManager.requestWayfindingUpdates(
                mWayfindingDestination!!,
                mWayfindingListener
            )
        }
    }

    override fun onPause() {
        // unregister location & region changes
        mIALocationManager.removeLocationUpdates(mListener)
        mIALocationManager.unregisterRegionListener(mRegionListener)
        mIALocationManager.unregisterOrientationListener(mOrientationListener)
        if (mWayfindingDestination != null) {
            mIALocationManager.removeWayfindingUpdates()
        }
    }

    override fun onDestroy() {
        mIALocationManager.destroy()
    }

    override val mOrientationListener: IAOrientationListener = object : IAOrientationListener {
        override fun onHeadingChanged(timestamp: Long, heading: Double) {
            updateHeading(heading)
        }

        override fun onOrientationChange(timestamp: Long, quaternion: DoubleArray) {
            // we do not need full device orientation in this example, just the heading
        }
    }

    override fun updateHeading(heading: Double) {
        // TODO
//        if (mHeadingMarker != null) {
//            mHeadingMarker!!.rotation = heading.toFloat()
//        }
    }

    override val mListener: IALocationListener = object : IALocationListenerSupport() {
        override fun onLocationChanged(location: IALocation) {
            Log.d(TAG, "new location with coordinates: ${location.latitude} ${location.longitude}")
            val center = IALatLng(location.latitude, location.longitude)
            if (mFloor != location.floorLevel) {
                updateRouteVisualization()
            }
            mFloor = location.floorLevel
            sendDataToFlutter(ChannelData.Location(location))
            if (cameraPositionNeedsUpdating) {
                cameraPositionNeedsUpdating = false
                sendDataToFlutter(ChannelData.UpdateCamera(center, 17.5f))
            }
        }
    }

    override val mRegionListener = object : IARegion.Listener {
        override fun onEnterRegion(region: IARegion) {
            if (region.type == IARegion.TYPE_FLOOR_PLAN) {
                Log.d(TAG, "Enter floor-plan ${region.id}")
                cameraPositionNeedsUpdating = true
                mCurrentFloorPlan = region
                mVenue.poIs
                fetchAndSetFloorPlan(region.floorPlan)   // or handleFloorPlanChange(region.floorPlan)
                setupPoIs(mVenue.poIs, region.floorPlan.floorLevel)
            } else if (region.type == IARegion.TYPE_VENUE) {
                mVenue = region.venue
            }
        }

        override fun onExitRegion(region: IARegion) {
            if (region.type == IARegion.TYPE_FLOOR_PLAN)
                mCurrentFloorPlan = null
        }

    }

    override val mWayfindingListener = IAWayfindingListener { route ->
        //mCurrentRoute = route
        if (hasArrivedToDestination(route)) { // stop wayfinding on reaching the destination
            // showInfo("You're there")
            mCurrentRoute = null
            mWayfindingDestination = null
            mIALocationManager.removeWayfindingUpdates()
        }
        updateRouteVisualization()
    }

    override fun updateRouteVisualization() {
        clearRouteVisualization()
        if (mCurrentRoute == null) {
            return
        }
        for (leg in mCurrentRoute!!.legs) {
            if (leg.edgeIndex == null) {
                continue
            }
            sendDataToFlutter(ChannelData.UpdateRouteVisualization(leg))
        }
    }

    /**
     * clear polylines from map
     */
    override fun clearRouteVisualization() {
        sendDataToFlutter(ChannelData.ClearPolylines)
    }

    override fun hasArrivedToDestination(route: IARoute): Boolean {
        if (route.legs.size == 0) return false

        val FINISH_THRESOLD_METERS = 8.0
        var routeLength = 0.0
        for (leg in route.legs) routeLength += leg.length
        return routeLength < FINISH_THRESOLD_METERS
    }

    override fun setWayfindingTarget(point: IALatLng, addMarker: Boolean) {
        mWayfindingDestination =
            IAWayfindingRequest.Builder()
                .withFloor(mFloor)
                .withLatitude(point.latitude)
                .withLongitude(point.longitude)
                .build()

        mIALocationManager
            .requestWayfindingUpdates(mWayfindingDestination!!, mWayfindingListener)
        sendDataToFlutter(ChannelData.Destination(IALatLng(point.latitude, point.longitude)))
        // Add destination marker on mapbox-flutter
    }

    override fun fetchAndSetFloorPlan(floorPlan: IAFloorPlan?) {
        floorPlan?.let { fp ->
            if (fp.id == mCurrentFloorPlan!!.id) {
                sendDataToFlutter(ChannelData.Image(floorPlan))
            }
        }
    }


    override fun setupPoIs(pois: List<IAPOI>, currentFloorLevel: Int) {
        Log.d(TAG, "${pois.size} PoI(s)")
        sendDataToFlutter(ChannelData.PoIs(pois, currentFloorLevel))
        // remove existing PoIs from mapbox-flutter
        // add current PoIs in mapbox-flutter
    }

    private var num = 0

    override fun ping(): String {
        num++
        return "Inside IAWayfinding $num"
    }

}