import 'package:flutter/services.dart';

import '../../utils/constants.dart';
import '../models/models.dart';

class IndoorAtlas {
  static bool initialized = false;
  static IAConfiguration opts = const IAConfiguration(apiKey: '');
  static IAConfiguration _opts = const IAConfiguration(apiKey: '');
  static IAConfiguration givenOpts = const IAConfiguration(apiKey: '');
  static String? _traceId;
  static final _listeners = Set<IAListener>.identity();
  static const bool _debugEnabled = false;
  static IAWayfindingRequest? _wayfinding;

  static IAVenue? _currentVenue;
  static IAFloorplan? _currentFloorplan;
  static IALocation? _currentLocation;
  static IARoute? _currentRoute;

  static Future<T?> _native<T>(String method, [dynamic arguments]) {
    return methodChannel.invokeMethod<T>(method, arguments);
  }

  static Future<T?> _nativeInitialized<T>(String method, [dynamic arguments]) {
    if (!initialized) {
      throw ('state inconsistency detected: indooratlas was not initialized before native call');
    }
    return methodChannel.invokeMethod<T>(method, arguments);
  }

  static void _debug(String msg) {
    if (true) print('IndoorAtlas DEBUG: $msg');
  }

  static void _warning(String msg) {
    print('IndoorAtlas WARNING: $msg');
  }

  static void _error(String msg) {
    print('IndoorAtlas ERROR: $msg');
    for (var l in _listeners) {
      l.onStatus(IAStatus.outOfService, msg);
    }
  }

  static Future<void> _onStatusChanged(int status) {
    // TODO: add informal messages?
    for (var l in _listeners) {
      l.onStatus(IAStatus.values[status], '');
    }
    return Future.value();
  }

  static Future<void> _onLocationChanged(Map location) {
    final loc = IALocation.fromMap(location);
    _currentLocation = loc;
    for (var l in _listeners) {
      l.onLocation(loc);
    }
    return Future.value();
  }

  static void _dispatchRegionEvent(bool enter, Region region) {
    if (region.venue != null) {
      _currentVenue = (enter ? region.venue : null);
      for (var l in _listeners) {
        l.onVenue(enter, region.venue!);
      }
    }
    if (region.floorplan != null) {
      _currentFloorplan = (enter ? region.floorplan : null);
      for (var l in _listeners) {
        l.onFloorplan(enter, region.floorplan!);
      }
    }
  }

  static Future<void> _onEnterRegion(Map region) {
    _dispatchRegionEvent(true, Region.fromMap(region));
    return Future.value();
  }

  static Future<void> _onExitRegion(Map region) {
    _dispatchRegionEvent(false, Region.fromMap(region));
    return Future.value();
  }

  static Future<void> _onOrientationChanged(
      int timestamp, double x, double y, double z, double w) {
    // for (var l in _listeners) {
    //   l.onOrientation(x, y, z, w);
    // }
    return Future.value();
  }

  static Future<void> _onHeadingChanged(int timestamp, double heading) {
    for (var l in _listeners) {
      l.onHeading(heading);
    }
    return Future.value();
  }

  static Future<void> _onWayfindingUpdate(Map route) {
    // should not happen, but check anyway
    if (_wayfinding != null) {
      final r = IARoute.fromMap(route);
      _currentRoute = r;
      if (_hasArrivedDestination(r)) {
        _currentRoute = null;
      }
      for (var l in _listeners) {
        l.onWayfindingRoute(r, _wayfinding);
      }
    }
    return Future.value();
  }

  static _hasArrivedDestination(IARoute route) {
    if (route.legs.isEmpty) {
      return false;
    }
    const finishThresoldMeters = 8.0;
    var routeLength = 0.0;
    for (var leg in route.legs) {
      routeLength += leg.length;
    }
    return routeLength < finishThresoldMeters;
  }

  //  Future<void> _onPermissionsGranted(bool granted) {
  //   if (!_permissions && granted && _initialized) {
  //     // reinitialize
  //     _initialized = false;
  //     _applyOptions(_opts);
  //   }
  //   _permissions = granted;
  //   return Future.value();
  // }

  static void _resetState() {
    if (_currentFloorplan != null) {
      for (var l in _listeners) {
        l.onFloorplan(false, _currentFloorplan!);
      }
      _currentFloorplan = null;
    }
    if (_currentVenue != null) {
      for (var l in _listeners) {
        l.onVenue(false, _currentVenue!);
      }
      _currentVenue = null;
    }
    _currentLocation = null;
  }

  static void _resetWayfinding() {
    if (_currentRoute != null) {
      for (var l in _listeners) {
        l.onWayfindingRoute(IARoute.empty(), null);
      }
      _currentRoute = null;
    }
  }

  static void _applyOptions(IAConfiguration opts) {
    if (!initialized) {
      methodChannel.setMethodCallHandler((call) {
        print('method call: $call');
        _nativeInitialized<String?>('getTraceId').then((v) => _traceId = v);
        try {
          switch (call.method) {
            case 'onStatusChanged':
              return Function.apply(_onStatusChanged, call.arguments);
            case 'onLocationChanged':
              return Function.apply(_onLocationChanged, call.arguments);
            case 'onEnterRegion':
              return Function.apply(_onEnterRegion, call.arguments);
            case 'onExitRegion':
              return Function.apply(_onExitRegion, call.arguments);
            case 'onOrientationChanged':
              return Function.apply(_onOrientationChanged, call.arguments);
            case 'onHeadingChanged':
              return Function.apply(_onHeadingChanged, call.arguments);
            case 'onWayfindingUpdate':
              return Function.apply(_onWayfindingUpdate, call.arguments);
            // case 'onPermissionsGranted':
            //   return Function.apply(_onPermissionsGranted, call.arguments);
          }
        } catch (err, stack) {
          _error('$err\n$stack');
        }
        return throw PlatformException(code: '1', message: 'Not Implemented');
      });
      _resetState();
      _resetWayfinding();
      _traceId = null;
      _native('initialize', ['0.0.1', opts.apiKey, opts.endpoint]);
      _debug('init done');
      initialized = true;
    }
    if (_listeners.isEmpty) {
      _debug('stopping positioning');
      _nativeInitialized('stopPositioning');
      _resetState();
    } else {
      _debug('starting positioning');
      if (opts.minChangeMeters != _opts.minChangeMeters) {
        _debug('minChangeMeters changed');
      }
      if (opts.minIntervalSeconds != _opts.minIntervalSeconds) {
        _debug('minIntervalSeconds changed');
      }
      _nativeInitialized('setOutputThresholds',
          [opts.minChangeMeters, opts.minIntervalSeconds]);
      if (opts.positioningMode != _opts.positioningMode) {
        _debug('positioningMode changed');
      }
      _nativeInitialized('setPositioningMode', [opts.positioningMode.index]);
      _nativeInitialized('startPositioning');
      if (opts.indoorLock != _opts.indoorLock) _debug('indoorLock changed');
      _nativeInitialized('lockIndoors', [opts.indoorLock]);
      if (opts.floorLock != _opts.floorLock) _debug('floorLock changed');
      if (opts.floorLock != null) {
        _nativeInitialized('lockFloor', [opts.floorLock!]);
      } else {
        _nativeInitialized('unlockFloor');
      }
    }

    if (_wayfinding == null) {
      _debug('stopping wayfinding');
      _nativeInitialized('stopWayfinding');
      _resetWayfinding();
    } else {
      _debug('starting wayfinding');
      _nativeInitialized('startWayfinding', [
        _wayfinding!.destination.latitude,
        _wayfinding!.destination.longitude,
        _wayfinding!.floor
      ]);
    }

    _opts = opts;
  }

  static void configure(IAConfiguration options) {
    givenOpts = options;
    _applyOptions(options);
  }

  static IAConfiguration get configuration => givenOpts;
  static String? get traceId => _traceId;
  static IAVenue? get venue => _currentVenue;
  static IAFloorplan? get floorplan => _currentFloorplan;
  static IALocation? get location => _currentLocation;

  static void setLocation(IACoordinate coordinate) {
    IALocation pos;
    if (coordinate is IALocation) {
      pos = coordinate;
    } else {
      pos = IALocation.fromCoordinate(coordinate, timestamp: DateTime.now());
    }
    _nativeInitialized(
        'setLocation', [pos.latitude, pos.longitude, pos.floor, pos.accuracy]);
  }

  static void subscribe(IAListener listener) {
    print("SUBSCRIBE LISTENER");
    if (_listeners.contains(listener)) return;
    _listeners.add(listener);

    // send current state
    if (_currentVenue != null) {
      listener.onVenue(true, _currentVenue!);
    }
    if (_currentFloorplan != null) {
      listener.onFloorplan(true, _currentFloorplan!);
    }
    if (_currentLocation != null) {
      listener.onLocation(_currentLocation!);
    }
    if (_wayfinding != null && _currentRoute != null) {
      listener.onWayfindingRoute(_currentRoute!, _wayfinding!);
    }

    _applyOptions(_opts);
  }

  static void unsubscribe(IAListener listener) {
    if (!_listeners.contains(listener)) return;
    _listeners.remove(listener);
    _applyOptions(_opts);
  }

  static void resubscribe(IAListener old, IAListener listener) {
    _listeners.remove(old);
    _listeners.add(listener);
  }

  // TODO: Get rid of the global wayfinding and use requestWayfindingRoute instead
  //       IAWayfindingRequest will be a IAListener member
  //       This way we get rid of some global state

  static void startWayfindingTo(IAWayfindingRequest request) {
    if (_wayfinding == request) return;
    _wayfinding = request;
    _applyOptions(_opts);
  }

  static void stopWayfindingTo(IAWayfindingRequest request) {
    if (_wayfinding == request) stopWayfinding();
  }

  static void stopWayfinding() {
    _wayfinding = null;
    _applyOptions(_opts);
  }
}
